/*
** Module   : UnpooledConnectionProvider.java
** Abstract : Simplified implementation of connection provider which does no
**            connection pooling
**
** Copyright (c) 2006-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20060619 @27501 Created initial version. Simple connection
**                         provider implementation designed to reduce
**                         complexity for debugging purposes.
** 002 SVL 20080725 @39175 Javadoc and logging output was modified
**                         in order to indicate that this class is
**                         used in production.
** 003 SVL 20120331        Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist;

import java.sql.*;
import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.util.*;
import org.hibernate.service.*;
import org.hibernate.service.jdbc.connections.spi.*;
import org.hibernate.service.spi.*;

/**
 * A simplified implementation of Hibernate's <code>ConnectionProvider</code>
 * interface which does no connection pooling, but instead fetches a new JDBC
 * connection each time {@link #getConnection} is invoked, and closes that
 * connection each time {@link #closeConnection} is invoked.
 * <p>
 * This implementation is used for in-memory databases, when connection
 * pooling actually works slower than opening a new connection each time. Also
 * it can be used for debugging purposes, it allows one to debug
 * connection-related issues while avoiding the behavioral side effects which
 * occur when a pooling connection provider recycles, proxies, caches, or
 * otherwise obfuscates JDBC connections.
 */
public final class UnpooledConnectionProvider
implements ConnectionProvider,
           Configurable
{
   /** Logger */
   private static final Log LOG =
      Utils.getLog(UnpooledConnectionProvider.class);
      
   /** URL to connect to the backing datastore */
   private String jdbcURL = null;
   
   /** User name used to connect */
   private String user = null;
   
   /** Password used to connect */
   private String password = null;
   
   /**
    * Default constructor required by Hibernate's connection provider
    * contract.
    */
   public UnpooledConnectionProvider()
   {
   }
   
   /**
    * Configure connection provider, extracting a URL, user name, and
    * password from the specified properties.
    *
    * @param   props
    *          Hibernate configuration properties.
    */
   public void configure(Map props)
   {
      jdbcURL = (String) props.get("hibernate.connection.url");
      user = (String) props.get("hibernate.connection.username");
      password = (String) props.get("hibernate.connection.password");

      if (LOG.isDebugEnabled())
      {
         LOG.debug("Simple connection provider configured for " + jdbcURL +
                   ";  user:  " + user + ";  password:  " +
                   (password == null ? "N/A" : "********"));
      }
   }
   
   /**
    * Obtain a new JDBC connection from the backing datastore.
    *
    * @return  New JDBC connection.
    */
   public Connection getConnection()
   throws SQLException
   {
      Connection conn = DriverManager.getConnection(jdbcURL, user, password);
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug("Getting JDBC connection:  " + conn);
      }
      
      return conn;
   }
   
   /**
    * Close the given JDBC connection.
    *
    * @param   conn
    *          Connection to close.
    */
   public void closeConnection(Connection conn)
   throws SQLException
   {
      if (LOG.isDebugEnabled())
      {
         LOG.debug("Closing JDBC connection:  " + conn);
      }
      
      conn.close();
   }
   
   /**
    * No-op implementation of required method.
    */
   public void close()
   {
   }
   
   /**
    * Returns <code>false</code> to indicate JDBC connections will not be
    * released aggressively.
    *
    * @return  <code>false</code>.
    */
   public boolean supportsAggressiveRelease()
   {
      return false;
   }

   /**
    * Can this wrapped service be unwrapped as the indicated type?
    *
    * @param  unwrapType
    *         The type to check.
    *
    * @return <code>true</code>/<code>false</code>
    */
   public boolean isUnwrappableAs(Class unwrapType)
   {
      return ConnectionProvider.class.equals( unwrapType ) ||
				UnpooledConnectionProvider.class.isAssignableFrom( unwrapType );
   }

   /**
    * Unproxy the service proxy.
    *
    * @param  unwrapType
    *         The java type as which to unwrap this instance.
    *
    * @return The unwrapped reference.
    *
    * @throws UnknownUnwrapTypeException if the service cannot be unwrapped as the indicated type.
    */
   public <T> T unwrap(Class<T> unwrapType)
   {
      if (isUnwrappableAs(unwrapType)) {
			return (T) this;
		}
		else {
			throw new UnknownUnwrapTypeException( unwrapType );
		}
   }
}
