/*
** Module   : H2Helper.java
** Abstract : Helper class which initializes embedded H2 database
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20080311   @37516 Created initial version. Helper class which
**                           initializes embedded H2 database.
** 002 ECF 20080320   @37649 Added new steps to database preparation. Set
**                           sorting of nulls to high (NULL values last).
**                           Set collation.
** 003 SVL 20080429   @38135 Collation is loaded from the directory.
** 004 ECF 20080508   @38487 Changed createFunctionAliases(). Use Database
**                           object when registering functions with the
**                           HQLPreprocessor rather than database name
**                           String. This allows differentiation of the
**                           same function across multiple databases with
**                           the same name.
** 005 ECF 20080724   @39169 Fixed bugs in executeDDL(). Connection and
**                           SessionFactory must not be closed after use.
** 006 ECF 20090729   @43451 Fixed logging in executeDDL(). It was possible to
**                           trigger NPE in BatchUpdateException catch block.
** 007 CA  20101029          Added method to retrieve a list of database 
**                           prepare statements.
** 009 CA  20101202          Added method to prepare a permanent database, as
**                           the custom functions need to  be registered with 
**                           the HQL preprocessor. Removed redundant code 
**                           (collation was set more than 1 time).
**                           The "h2.sortNullsHigh" setting was removed, as
**                           it wasn't taken into consideration anyway; also,
**                           more testing needs to be done to exactly 
**                           determine how 4GL sorts the unknown value in the
**                           temporary and permanent database.
** 010 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.embed;

import java.lang.reflect.*;
import java.sql.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.pl.*;
import com.goldencode.p2j.util.*;
import org.hibernate.internal.*;

/**
 * Helper class which initializes an embedded H2 database.
 */
public final class H2Helper
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(H2Helper.class);
   
   /**
    * Private constructor.  All access to this class is via static methods.
    */
   private H2Helper()
   {
   }
   
   /**
    * Prepare the specified database by creating all necessary function
    * aliases and setting global parameters.
    * <p>
    * The "h2.sortNullsHigh" setting was removed, as the H2 driver was already
    * initiated at the time when this method is called, so the setting has no
    * effect.
    * 
    * @param   database
    *          Database to prepare.
    *          
    * @throws  PersistenceException
    *          if there is any error executing DDL.
    */
   public static void prepareDatabase(Database database)
   throws PersistenceException
   {
      // Create built-in function aliases.
      createFunctionAliases(database,
                            BuiltIns.getPublicFunctions(),
                            true);
      createFunctionAliases(database,
                            BuiltIns.getSupportFunctions(),
                            false);
      
      // Set collation.
      String path = String.format("database/%s/p2j/embedded-collation",
                                  database.getName());

      String collation = Utils.getDirectoryNodeString(null,
                                                      path,
                                                      null,
                                                      false);
      if (collation != null)
      {
         try
         {
            List<String> ddl = new ArrayList<String>(1);
            ddl.add(String.format("set collation \"%s\"", collation));
            executeDDL(ddl, database);
         }
         catch (PersistenceException exc)
         {
            LOG.error(
               "Error setting P2J-specific collation;  please ensure " +
               "p2jspi.jar is installed in Java extension directory");

            throw exc;
         }
      }
      else if (LOG.isWarnEnabled())
      {
         StringBuilder buf = new StringBuilder();
         buf.append("Embedded ");
         buf.append(database.isDirty() ? "dirty" : "primary");
         buf.append(" database '");
         buf.append(database.getName());
         buf.append("' collation not set;  using default Java collation");
         LOG.warn(buf);
      }
   }
   
   /**
    * Register the function aliases with the HQL preprocessor, for the 
    * permanent database. 
    * 
    * @param   database
    *          Database to prepare.
    */
   public static void preparePermanentDatabase(Database database)
   {
      getFunctionAliases(database, 
                         BuiltIns.getPublicFunctions(), 
                         true, true);
      getFunctionAliases(database, 
                         BuiltIns.getSupportFunctions(), 
                         false, true);
   }
   
   /**
    * Get a list of database prepare DDL statements.
    * 
    * @param   database
    *          Database to prepare.
    *          
    * @return  the list of DDL statements.
    */
   public static List<String> getPrepareStatements(Database database)
   {
      List<String> ddl = new ArrayList<String>();
      
      // Create built-in function aliases.
      ddl.addAll(getFunctionAliases(database,
                                    BuiltIns.getPublicFunctions(),
                                    true, false));
      ddl.addAll(getFunctionAliases(database,
                                    BuiltIns.getSupportFunctions(),
                                    false, false));
      
      return ddl;
   }
   
   /**
    * Get function aliases for the given list of static methods, optionally
    * making the alias names unique by appending a unique number for each
    * overloaded method.
    * 
    * @param   database
    *          Database in which to create aliases.
    * @param   funcs
    *          List of static methods for which function aliases should be
    *          created.
    * @param   overload
    *          <code>true</code> to create unique function names for each
    *          overloaded method
    * @param   register
    *          <code>true</code> if the function should be registered with
    *          the {@link HQLPreprocessor}.
    *          
    * @throws  PersistenceException
    *          if there is any error executing DDL.
    */
   private static List<String> getFunctionAliases(Database database,
                                                  List<Method> funcs,
                                                  boolean overload,
                                                  boolean register)
   {
      List<String> ddl = new ArrayList<String>();
      Map<String, Integer> overloads = new HashMap<String, Integer>();
      Iterator<Method> iter = funcs.iterator();
      while (iter.hasNext())
      {
         Method method = iter.next();
         String name = method.getName();
         Integer count = null;
         if (overload)
         {
            count = overloads.get(name);
            if (count == null)
            {
               count = 1;
            }
            else
            {
               count++;
            }
            overloads.put(name, count);
         }
         
         StringBuilder buf = new StringBuilder();
         
         buf.append(name);
         
         if (overload)
         {
            buf.append('_');
            buf.append(count);
         }
         String uniqueName = buf.toString();
         
         buf.insert(0, "create alias if not exists ");
         buf.append(" for ");
         buf.append(formatMethod(method));
         buf.append(";");

         ddl.add(buf.toString());
         
         if (overload && register)
         {
            HQLPreprocessor.registerFunction(database, uniqueName, method);
         }
      }
      
      return ddl;
   }
   
   /**
    * Create function aliases for the given list of static methods, optionally
    * making the alias names unique by appending a unique number for each
    * overloaded method.
    * 
    * @param   database
    *          Database in which to create aliases.
    * @param   funcs
    *          List of static methods for which function aliases should be
    *          created.
    * @param   overload
    *          <code>true</code> to create unique function names for each
    *          overloaded method
    * 
    * @throws  PersistenceException
    *          if there is any error executing DDL.
    */
   private static void createFunctionAliases(Database database,
                                             List<Method> funcs,
                                             boolean overload)
   throws PersistenceException
   {
      List<String> ddl = getFunctionAliases(database, funcs, overload, true);
      
      executeDDL(ddl, database);
   }
   
   /**
    * Generate a quoted method name and signature, suitable for use within a
    * CREATE ALIAS DDL statement.
    * 
    * @param   method
    *          Method which is to be mapped to a function alias.
    * 
    * @return  Quoted method name and signature.
    */
   private static String formatMethod(Method method)
   {
      StringBuilder buf = new StringBuilder();
      buf.append("\"");
      buf.append(method.getDeclaringClass().getName());
      buf.append('.');
      buf.append(method.getName());
      buf.append('(');
      Class<?>[] sig = method.getParameterTypes();
      for (int i = 0; i < sig.length; i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         
         buf.append(sig[i].getName());
      }
      buf.append(')');
      buf.append("\"");
      
      return buf.toString();
   }
   
   /**
    * Execute a batch of DDL statements for the given database.  The batch is
    * executed within a single transaction.
    * 
    * @param ddl
    * @param database
    * @throws PersistenceException
    */
   private static void executeDDL(List<String> ddl, Database database)
   throws PersistenceException
   {
      SessionFactory factory = DatabaseManager.getSessionFactory(database);
      Session session = factory.openSession();
      Connection conn = ((SessionImpl) session).connection();
      Statement statement = null;
      try
      {
         try
         {
            boolean debug = LOG.isDebugEnabled();
            conn.setAutoCommit(false);
            statement = conn.createStatement();
            for (String next : ddl)
            {
               statement.addBatch(next);
               if (debug)
               {
                  LOG.debug("DDL:  " + next);
               }
            }
            
            statement.executeBatch();
            conn.commit();
         }
         catch (BatchUpdateException exc)
         {
            conn.rollback();
            
            SQLException next = exc.getNextException();
            
            throw new PersistenceException(
               "Error initializing embedded database",
               (next != null ? next : exc));
         }
         catch (SQLException exc)
         {
            conn.rollback();
            
            throw new PersistenceException(
               "Error initializing embedded database", exc);
         }
         finally
         {
            if (statement != null)
            {
               statement.close();
            }
            
            if (session != null && session.isOpen())
            {
               session.close();
            }
         }
      }
      catch (SQLException exc)
      {
      }
   }
}
