/*
** Module   : JdbcKeyRetriever.java
** Abstract : KeyRetriever implementation which retrieves keys using SQL
**            queries.
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 SVL 20080827  @39779 Created initial version. Retrieves free keys
**                          using two subsequent queries.
** 002 SVL 20091126  @44444 Use external function for retrieving the list of
**                          database tables.
** 003 SVL 20120331         Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.id;

import com.goldencode.p2j.persist.*;
import java.util.*;
import org.hibernate.*;
import org.hibernate.type.*;


/**
 * Provides a service for retrieving available primary keys from database
 * using SQL queries.
 */
public class JdbcKeyRetriever
implements KeyRetriever
{
   /** Query which returns the number of busy keys in the given range. */
   private String predictionQuery;

   /** Query which retrieves all busy keys in the given range. */
   private String retrievalQuery;

   /**
    * Initialize key retriever for the given database.
    *
    * @param database
    *        Database for which will be scanned for available keys.
    */
   public void initialize(Database database)
   {
      // Get all table names.
      List<String> tableNames = DatabaseManager.getDatabaseTables(database);

      // Compose prediction query which looks like
      // 'select(
      //  (select count(id) from table1 where id >= :minId and id <= :maxId) +
      //  (select count(id) from table2 where id >= :minId and id <= :maxId) +
      //  ...) as cnt'
      StringBuilder buf = new StringBuilder("select (");

      for (int i = 0; i < tableNames.size(); i++)
      {
         String tableName = tableNames.get(i);

         buf.append("(select count(id) from ");
         buf.append(tableName);
         buf.append(" where id >= :minId and id <= :maxId)");

         if (i != tableNames.size() - 1)
         {
            buf.append(" + ");
         }
      }
      buf.append(") as cnt");

      predictionQuery = buf.toString();

      // Compose id retrieval query which looks like
      // 'select id from table1 where id >= :minId and id <= :maxId UNION
      //  select id from table2 where id >= :minId and id <= :maxId UNION
      //  ....  ORDER BY id asc'
      buf = new StringBuilder();

      for (int i = 0; i < tableNames.size(); i++)
      {
         String tableName = tableNames.get(i);

         buf.append("select id from ");
         buf.append(tableName);
         buf.append(" where id >= :minId and id <= :maxId");

         if (i != tableNames.size() - 1)
         {
            buf.append(" UNION ");
         }
      }
      buf.append(" ORDER BY id asc");

      retrievalQuery = buf.toString();
   }

   /**
    * Get available keys.
    *
    * @param  minKey
    *         Lower bound of a range of identifiers on which search will be
    *         made.
    * @param  maxKey
    *         Upper bound of a range of identifiers on which search will be
    *         made.
    * @param  session
    *         Hibernate session which will be used to perform queries.
    * @return A list of ranges of available keys
    */
   public List<Range> retrieveFreeKeys(Long minKey,
                                       Long maxKey,
                                       Session session)
   {
      Long rangeSize = maxKey - minKey + 1;

      // Execute prediction query.
      SQLQuery query = session.createSQLQuery(predictionQuery);
      query.addScalar("cnt", StandardBasicTypes.LONG);
      query.setLong("minId", minKey);
      query.setLong("maxId", maxKey);

      Long count = (Long)query.uniqueResult();

      if (count.equals(0L))
      {
         // All ids in the given range are free.
         List<Range> res = new ArrayList<Range>();
         res.add(new Range(minKey, maxKey));
         return res;
      }
      else if (count.equals(rangeSize))
      {
         // No ids available.
         return null;
      }

      // Execute the query which will return all busy ids in the given range.
      query = session.createSQLQuery(retrievalQuery);
      query.addScalar("id", StandardBasicTypes.LONG);
      query.setLong("minId", minKey);
      query.setLong("maxId", maxKey);

      List<Long> busyKeys = (List<Long>) query.list();
      
      return new Range(minKey, maxKey).splitRange(busyKeys);
   }
}
