/*
** Module   : StoredProcKeyRetriever.java
** Abstract : KeyRetriever implementation which retrieves keys using stored
**            procedure.
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVL 20080827   @39782 Created initial version. Retrieves free keys
**                           using a stored procedure.
** 002 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.id;

import com.goldencode.p2j.persist.Database;
import java.util.*;
import org.hibernate.*;
import org.hibernate.type.*;

/**
 * Provides a service for retrieving available primary keys from database
 * by calling stored procedure.
 */
public class StoredProcKeyRetriever
implements KeyRetriever
{
   /** Query which is used for stored procedure call. */
   private String retrieveQuery;

   /**
    * Name of the stored procedure used for id retrieval.
    * Procedure should be declared as 'FUNCTION procedure_name(min_id bigint, 
    * max_id bigint) RETURNS SETOF range_type'
    */
   private static String STORED_PRUCEDURE_NAME = "get_free_ids";

   /**
    * Initialize key retriever for the given database.
    *
    * @param database
    *        Database for which will be scanned for available keys.
    */
   public void initialize(Database database)
   {
      StringBuilder buf = new StringBuilder();
      buf.append("select lower_bound, upper_bound from ");
      buf.append(STORED_PRUCEDURE_NAME);
      buf.append("(:minId, :maxId)");

      retrieveQuery = buf.toString();
   }

   /**
    * Get available keys.
    *
    * @param  minKey
    *         Lower bound of a range of identifiers on which search will be
    *         made.
    * @param  maxKey
    *         Upper bound of a range of identifiers on which search will be
    *         made.
    * @param  session
    *         Hibernate session which will be used to perform queries.
    * @return A list of ranges of available keys
    */
   public List<Range> retrieveFreeKeys(Long minKey,
                                       Long maxKey,
                                       Session session)
   {
      // Execute the query which will return all ranges of free ids in the
      // given global range. 
      SQLQuery query = session.createSQLQuery(retrieveQuery);
      query.addScalar("lower_bound", StandardBasicTypes.LONG);
      query.addScalar("upper_bound", StandardBasicTypes.LONG);
      query.setLong("minId", minKey);
      query.setLong("maxId", maxKey);

      List<Range> res = new ArrayList<Range>();
      List<Object[]> bounds = (List<Object[]>) query.list();
      for (Object[] bound : bounds)
      {
         res.add(new Range((Long)bound[0], (Long)bound[1]));
      }

      return res;
   }
}
