/*
** Module   : IntegerUserType.java
** Abstract : Custom user type to wrap P2J integer for Hibernate access
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20050630   @21590 Created initial version. Maps P2J integer to
**                           SQL integer.
** 002 ECF 20050729   @21955 Refactored to move nullSafeSet method
**                           implementation up to superclass. This was
**                           necessary to add centralized logging for
**                           prepared statement parameter binding.
** 003 ECF 20050809   @22075 Fixed nullSafeGet and added logging. Defect
**                           was that method was checking whether the
**                           result set column was null *before* reading
**                           data at that column. Also made class final.
** 004 ECF 20060907   @29316 Fixed potential ClassCastException in
**                           bindNonNull() and getString() methods. Must
**                           cast to NumberType instead of integer, and
**                           use intValue() instead of getValue().
** 005 ECF 20060907   @29332 Rewrote deepCopy(). Needed to handle unknown
**                           value safely.
** 006 ECF 20080806   @39298 Moved to type sub-package of persist.
** 007 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.integer P2J integer}. This permits Hibernate to
 * convert this data type to and from the SQL INTEGER data type. The unknown
 * value is mapped to SQL NULL.
 */
public final class IntegerUserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { Types.INTEGER };
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map an integer to/from a database
    * record.
    *
    * @return  Array containing a single type constant:  INTEGER.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.integer P2J integer} class.
    */
   public Class returnedClass()
   {
      return integer.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>integer</code>
    * instance.
    *
    * @param   value
    *          The <code>integer</code> instance to be copied.
    *
    * @return  A new instance of <code>integer</code> which contains the
    *          same internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      NumberType num = (NumberType) value;
      if (num.isUnknown())
      {
         return (new integer());
      }
      
      return (new integer(num.intValue()));
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon
    * a data value in the specified result set. If the corresponding data
    * value is NULL in the result set, return an <code>integer</code>
    * instance which is initialized to the unknown value.
    *
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          an <code>int</code>.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  An <code>integer</code> instance, possibly initialized to the
    *          unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs, String[] names, Object owner)
   throws SQLException
   {
      int value = rs.getInt(names[0]);
      
      if (rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.trace("returning unknown value as column: " + names[0]);
         }
         return new integer();
      }
      
      integer wrapper = new integer(value);
      if (LOG_TRACE)
      {
         LOG.trace("returning '" + toString(wrapper) + "' as column: " +
                   names[0]);
      }
      
      return wrapper;
   }
   
   /**
    * Set an <code>int</code> value into the given prepared statement,
    * based upon the <code>integer</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          An <code>integer</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps,
                              BaseDataType value,
                              int index)
   throws SQLException
   {
      ps.setInt(index, ((NumberType) value).intValue());
   }
   
   /**
    * Return the string value of the <code>int</code> embedded in the
    * given <code>integer</code> object <code>value</code>. The default
    * <code>toString()</code> method of <code>integer</code> returns a
    * formatted, padded string, which is not suitable for our purposes.
    *
    * @param   value
    *          An <code>integer</code> object. It is safe to assume this
    *          will not be <code>null</code>.
    *
    * @return  The <code>String</code> representation of the contents of
    *          <code>value</code>.
    */
   protected String toString(BaseDataType value)
   {
      return String.valueOf(((NumberType) value).intValue());
   }

   /**
    * Extract a data value (not wrapped by a <code>BaseDataType</code>
    * implementation) object from a result set at a particular column.
    *
    * @param   rs
    *          Result set containing data to be extracted.
    * @param   column
    *          Name of the column at which data will be extracted.
    *
    * @return  An object which will be wrapped inside a P2J wrapper object,
    *          or <code>null</code> if the SQL value was NULL at the target
    *          column.
    *
    * @throws  SQLException
    *          if a database access error occurs.
    */
   protected Object getBaseValue(ResultSet rs, String column)
   throws SQLException
   {
      return rs.getInt(column);
   }

   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which wraps <code>value</code>.
    *
    * @param   value
    *          Value to be wrapped by a concrete subclass of
    *          <code>BaseDataType<code>; will not be <code>null</code>.
    *
    * @return  Instance of the appropriate <code>BaseDataType<code> concrete
    *          implementation.
    */
   protected final BaseDataType instantiateRegular(Object value)
   {
      return new integer((Integer) value);
   }

   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which is initialized to the unknown value.
    *
    * @return  Data wrapper object initialized to unknown value.
    */
   protected final BaseDataType instantiateUnknown()
   {
      return new integer();
   }
}
