/*
** Module   : RawUserType.java
** Abstract : Custom user type to wrap P2J raw for Hibernate access
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20050908   @22602 Created initial version. Maps P2J raw to SQL
**                           byte array. Replaces Hibernate binary type as
**                           the mapping for Progress raw values.
** 002 ECF 20051025   @23169 Added isMutable method. Overrides superclass'
**                           implementation to return true, reflecting the
**                           DMO access strategy for raw wrapper user
**                           types.
** 003 ECF 20080806   @39300 Moved to type sub-package of persist.
** 004 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.raw P2J raw} data type. This permits Hibernate to
 * convert this data type to and from the SQL VARBINARY data type. The
 * unknown value is mapped to SQL NULL.
 */
public final class RawUserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { Types.VARBINARY };
   
   /**
    * Returns <code>true</code> to indicate that the P2J data wrapper object
    * is mutable.  DMOs which get and set these objects create a non-final
    * reference.  Getters return the reference directly, and setters replace
    * the reference with that of the object provided by the caller.  Hence,
    * changes to the DMO's state can be made by changing the wrapper's state
    * directly, necessitating dirty checking by Hibernate.
    *
    * @return  <code>true</code>.
    */
   public boolean isMutable()
   {
      return true;
   }
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map a raw value to/from a
    * database record.
    *
    * @return  Array containing a single type constant:  VARBINARY.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.raw P2J raw} class.
    */
   public Class returnedClass()
   {
      return raw.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>raw</code> instance.
    *
    * @param   value
    *          The <code>raw</code> instance to be copied.
    *
    * @return  A new instance of <code>raw</code> which contains the same
    *          internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      return ((raw) value).duplicate();
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon
    * a data value in the specified result set. If the corresponding data
    * value is NULL in the result set, return a <code>raw</code> instance
    * which is initialized to the unknown value.
    *
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          a <code>byte</code> array.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  A <code>raw</code> instance, possibly initialized to the
    *          unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs, String[] names, Object owner)
   throws SQLException
   {
      byte[] value = rs.getBytes(names[0]);
      raw wrapper = null;
      
      if (rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.trace("returning unknown value as column: " + names[0]);
         }
         wrapper = new raw();
         wrapper.setUnknown();
      }
      else
      {
         wrapper = new raw(value);
         if (LOG_TRACE)
         {
            LOG.trace("returning '" + toString(wrapper) + "' as column: " +
                      names[0]);
         }
      }
      
      return wrapper;
   }
   
   /**
    * Set a <code>byte</code> array value into the given prepared statement,
    * based upon the <code>raw</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          A <code>raw</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException    
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps,
                              BaseDataType value,
                              int index)
   throws SQLException
   {
      ps.setBytes(index, ((raw) value).getValue());
   }

   /**
    * Extract a data value (not wrapped by a <code>BaseDataType</code>
    * implementation) object from a result set at a particular column.
    *
    * @param   rs
    *          Result set containing data to be extracted.
    * @param   column
    *          Name of the column at which data will be extracted.
    *
    * @return  An object which will be wrapped inside a P2J wrapper object,
    *          or <code>null</code> if the SQL value was NULL at the target
    *          column.
    *
    * @throws  SQLException
    *          if a database access error occurs.
    */
   protected Object getBaseValue(ResultSet rs, String column)
   throws SQLException
   {
      return rs.getBytes(column);
   }

   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which wraps <code>value</code>.
    *
    * @param   value
    *          Value to be wrapped by a concrete subclass of
    *          <code>BaseDataType<code>; will not be <code>null</code>.
    *
    * @return  Instance of the appropriate <code>BaseDataType<code> concrete
    *          implementation.
    */
   protected final BaseDataType instantiateRegular(Object value)
   {
      return new raw((byte[]) value);
   }

   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which is initialized to the unknown value.
    *
    * @return  Data wrapper object initialized to unknown value.
    */
   protected final BaseDataType instantiateUnknown()
   {
      return new raw();
   }
}
