/*
** Module   : RowIDUserType.java
** Abstract : Custom user type to wrap P2J rowid for Hibernate access
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20080731   @39293 Created initial version. Maps P2J rowid to
**                           SQL BIGINT.
** 002 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a custom, hibernate user type for the {@link
 * com.goldencode.p2j.util.rowid P2J rowid}. This permits Hibernate to
 * convert this data type to and from the SQL BIGINT data type. The unknown
 * value is mapped to SQL NULL.
 */
public final class RowIDUserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { Types.BIGINT };
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map an rowid to/from a database
    * record.
    *
    * @return  Array containing a single type constant:  BIGINT.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.rowid P2J rowid} class.
    */
   public Class returnedClass()
   {
      return rowid.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>rowid</code>
    * instance.
    *
    * @param   value
    *          The <code>rowid</code> instance to be copied.
    *
    * @return  A new instance of <code>rowid</code> which contains the
    *          same internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      rowid id = (rowid) value;
      
      return (new rowid(id));
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon a
    * data value in the specified result set. If the corresponding data value
    * is NULL in the result set, return a <code>rowid</code> instance which is
    * initialized to the unknown value.
    * 
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          a <code>long</code>.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  A <code>rowid</code> instance, possibly initialized to the
    *          unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs, String[] names, Object owner)
   throws SQLException
   {
      long value = rs.getLong(names[0]);
      
      if (rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.trace("returning unknown value as column: " + names[0]);
         }
         return new rowid();
      }
      
      rowid wrapper = new rowid(value);
      if (LOG_TRACE)
      {
         LOG.trace("returning '" + toString(wrapper) + "' as column: " +
                   names[0]);
      }
      
      return wrapper;
   }
   
   /**
    * Set a <code>long</code> value into the given prepared statement, based
    * upon the <code>rowid</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          A <code>rowid</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps,
                              BaseDataType value,
                              int index)
   throws SQLException
   {
      ps.setLong(index, ((rowid) value).getValue());
   }
   
   /**
    * Return the string value of the <code>long</code> embedded in the given
    * <code>rowid</code> object <code>value</code>. The default
    * <code>toString()</code> method of <code>rowid</code> returns a
    * formatted, padded string, which is not suitable for our purposes.
    *
    * @param   value
    *          A <code>rowid</code> object. It is safe to assume this will not
    *          be <code>null</code>.
    *
    * @return  The <code>String</code> representation of the contents of
    *          <code>value</code>.
    */
   protected String toString(BaseDataType value)
   {
      return String.valueOf(((rowid) value).getValue());
   }

   /**
    * Extract a data value (not wrapped by a <code>BaseDataType</code>
    * implementation) object from a result set at a particular column.
    *
    * @param   rs
    *          Result set containing data to be extracted.
    * @param   column
    *          Name of the column at which data will be extracted.
    *
    * @return  An object which will be wrapped inside a P2J wrapper object,
    *          or <code>null</code> if the SQL value was NULL at the target
    *          column.
    *
    * @throws  SQLException
    *          if a database access error occurs.
    */
   protected Object getBaseValue(ResultSet rs, String column)
   throws SQLException
   {
      return rs.getLong(column);
   }

   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which wraps <code>value</code>.
    *
    * @param   value
    *          Value to be wrapped by a concrete subclass of
    *          <code>BaseDataType<code>; will not be <code>null</code>.
    *
    * @return  Instance of the appropriate <code>BaseDataType<code> concrete
    *          implementation.
    */
   protected final BaseDataType instantiateRegular(Object value)
   {
      return new rowid((Long) value);
   }

   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which is initialized to the unknown value.
    *
    * @return  Data wrapper object initialized to unknown value.
    */
   protected final BaseDataType instantiateUnknown()
   {
      return new rowid();
   }
}
